
#include "base.as"

#library    lSets    "Sets.ao"
import from lSets

+++ 
+++  A Set is a FiniteSet if it is isomorphic to a (necessarily finite) Number object
+++
define FiniteSet:Category == Set with CountablyFinite

+++
+++  Forgetful functor to Set
+++
ForgetFiniteSet(A:FiniteSet):Set == A add
ForgetFiniteSet(A:FiniteSet,B:FiniteSet,f:A->B):ForgetFiniteSet A -> ForgetFiniteSet B == 
    f pretend (ForgetFiniteSet A -> ForgetFiniteSet B)

+++
+++  A handy domain constructor making a FiniteSet from a list of elements
+++  of a Set.  Finiteset(S,{S1,S2,..,SN}) is the finite set {S1,S2,...,SN}
+++    
FiniteSet(S:Set,l:List S):FiniteSet == add
    Rep == S; import from Rep
    (x:%)=(y:%):Boolean == rep x = rep y
    elements():Generator % == generate for s in l repeat yield per s
    (t:TextWriter)<<(x:%):TextWriter == t << rep x
    
define FiniteSetCategory:MathCategory FiniteSet with _
                                Final FiniteSet with _
                              Initial FiniteSet with _
                              Product FiniteSet with _
                            CoProduct FiniteSet with _
                             Quotient FiniteSet with _
                            Subobject FiniteSet with _
                            Equalizer FiniteSet with _
                          CoEqualizer FiniteSet with _
                             Pullback FiniteSet with _
                          Exponential FiniteSet with _
                          {  One: () -> FiniteSet }
== add
--      this causes hash value read on incomplete domain
--
--    1:FiniteSet == 
--        FS0:FiniteSet == (1$SetCategory) add
--            elements:Generator % == generate { yield (1$SingleInteger) pretend % }
--        FS0 add
--
--      this doesn't seem to help either... same error: see FiniteSetsTest.as
--
    One():FiniteSet == 
         FS0:FiniteSet == add
             Rep == SingleInteger; import from Rep
             (t:TextWriter)<<(x:%):TextWriter == t << "*"
             (x:%)=(y:%):Boolean == true
             elements(): Generator % == generate yield (1$SingleInteger) pretend %
         FS0 add
    one(A:FiniteSet):(A->One()) == (a:A):One() +-> (1$SingleInteger) pretend One()
    Zero():FiniteSet == 
        FS1:FiniteSet == (One()$SetCategory) add
            elements():Generator % == generate { }
        FS1 add
    zero(A:FiniteSet):(Zero()->A) == null(Zero(),A)
    Product(A:FiniteSet,B:FiniteSet):(AB:FiniteSet,AB->A,AB->B,(X:FiniteSet)->(X->A,X->B)->(X->AB)) ==
        RepP == Record(a:A,b:B); import from RepP
	P:FiniteSet == 
	    W2:FiniteSet == add
	        Rep == Record(a:A,b:B); import from RepP
		(x:%)=(y:%):Boolean == (rep x).a = (rep y).a and (rep x).b = (rep y).b
		elements():Generator % == generate
		    for aa:A in elements() for bb:B in elements() repeat yield per [aa,bb]
                (t:TextWriter)<<(x:%):TextWriter == t << "( " << (rep x).a << ", " << (rep x).b << " )"
	    W2 add 
	pa:P->A == (p:P):A +-> (p pretend RepP).a
	pb:P->B == (p:P):B +-> (p pretend RepP).b
	product(X:FiniteSet)(f:X->A,g:X->B):(X->P) == (x:X):P +-> [f x,g x] pretend P
	(P,pa,pb,product)
	
    CoProduct(A:FiniteSet,B:FiniteSet):(AB:FiniteSet,A->AB,B->AB,(X:FiniteSet)->(A->X,B->X)->(AB->X)) == 
        RepS == Union(a:A,b:B); import from RepS
	S:FiniteSet == 
	    W3:FiniteSet == add
	        Rep == Union(a:A,b:B); import from Rep
                (x:%)=(y:%):Boolean == 
                    (rep x) case a and (rep y) case a => (rep x).a = (rep y).a
                    (rep x) case a and (rep y) case b => false
                    (rep x) case b and (rep y) case a => false
                    (rep x) case b and (rep y) case b => (rep x).b = (rep y).b
                    never
		elements():Generator % == generate
		    for aa:A in elements() repeat yield [aa] pretend %
		    for bb:B in elements() repeat yield [bb] pretend %
                (t:TextWriter)<<(x:%):TextWriter == 
                    (rep x) case a => t << (rep x).a
                    (rep x) case b => t << (rep x).b
                    never
            W3 add
	inserta:A->S == (a:A):S +-> [a] pretend S
	insertb:B->S == (b:B):S +-> [b] pretend S
	sum(X:FiniteSet)(f:A->X,g:B->X):(S->X) == (s:S):X +-> { z:RepS == s pretend RepS; {z case a => f z.a; z case b => g z.b; never } }
        (S,inserta,insertb,sum)
	
    Quotient(A:FiniteSet,B:FiniteSet)(epic:A->B):(Quo:FiniteSet,A->Quo,(X:FiniteSet)->(A->X)->(Quo->X)) == 
        Q:FiniteSet == 
	    W4:FiniteSet == add
	        Rep == A; import from Rep,B
	        (x:%)=(y:%):Boolean == (rep x)=(rep y) or epic rep x = epic rep y
		elements():Generator % == generate for aa:A in elements() repeat yield per aa
                (t:TextWriter)<<(x:%):TextWriter == t << rep x
	    W4 add
        insert:A->Q == (a:A):Q +-> a pretend Q
	quotient(X:FiniteSet)(respectful:A->X):(Q->X) == (q:Q):X +-> respectful (q pretend A)
	(Q,insert,quotient)
	
    Subobject(A:FiniteSet,B:FiniteSet)(monic:B->A):(Sub:FiniteSet,Sub->A,(X:FiniteSet)->(X->A)->(X->Sub)) ==
        S:FiniteSet == 
	    W5:FiniteSet == add
	        Rep == A; import from Rep,B
		(x:%)=(y:%):Boolean == rep x = rep y
		elements():Generator % == generate for bb:B in elements() repeat yield per monic bb
                (t:TextWriter)<<(x:%):TextWriter == t << rep x
	    W5 add
	include:S->A == (s:S):A +-> s pretend A
	restrict(X:Set)(respectful:X->A):(X->S) == (x:X):S +-> (respectful x) pretend S 
	(S,include,restrict)
	
    Equalizer(A:FiniteSet,B:FiniteSet,f:A->B,g:A->B):(E:FiniteSet,E->A) == 
        import from List A,B
        Eq:FiniteSet == FiniteSet(A,[a for a in (elements$A)() | f a = g a])
        e:Eq->A == (x:Eq):A +-> x pretend A
	(Eq,e)

    CoEqualizer(A:FiniteSet,B:FiniteSet,f:B->A,g:B->A):(E:FiniteSet,A->E) == 
        CoEq:FiniteSet == 
	    W6:FiniteSet == add
	        Rep == A; import from Rep,B,lBasics
	        (x:%)=(y:%):Boolean == rep x = rep y or exists? ( f b = (rep x) and g b = (rep y)                for b:B in elements() )
		elements():Generator % == generate for a:A in elements() repeat yield per a
                (t:TextWriter)<<(x:%):TextWriter == t << rep x
	    W6 add
	coeq:A->CoEq == (a:A):CoEq +-> a pretend CoEq
	(CoEq,coeq)
	
    Exp(E:FiniteSet):Exp(FiniteSet,E) == 
        W8:Exp(FiniteSet,E) == add
            rightProductFunctor(A:FiniteSet):FiniteSet == 
                W9:FiniteSet == (A * E) add
                W9 add
            expFunctor(A:FiniteSet):FiniteSet == 
                W7:FiniteSet == add
                    Rep ==> E->A; import from A
                    (t:TextWriter)<<(x:%):TextWriter == 
                        pr(w:TextWriter,r:Record(d:E,c:A)):TextWriter == w << r.d << "+-> " << r.c
                        l:List Record(d:E,c:A) == [ [e,(rep x) e] for e:E in (elements$E)() ]
                        (ww:TextWriter)**(l:List Record(d:E,c:A)):TextWriter == 
                            empty? l => ww
                            (pr(ww,first l) << newline) ** rest l
                        t ** l
                    elements():Generator % == error "Write me"
                    (f:%)=(g:%):Boolean == forall?( (rep f) e = (rep g) e for e:E in (elements$E)() )
                W7 add
            >>(A:FiniteSet,B:FiniteSet,pr:rightProductFunctor A -> B):(A->expFunctor B) ==
                ff(aa:A):expFunctor B ==
                    ee(e:E):B ==  
                        r:Record(a:A,b:E) == [aa,e]
                        pr ( r pretend rightProductFunctor A )
                    ee pretend expFunctor B 
                ff 
            <<(A:FiniteSet,B:FiniteSet,f:A->expFunctor B):rightProductFunctor A -> B == 
                ff(x:rightProductFunctor A):B == 
                    r:Record(a:A,b:E) == x pretend Record(a:A,b:E)
                    ee:E->B == (f (r.a)) pretend E->B
                    ee ( r.b )
                ff
        W8 add    

    Pullback(A:FiniteSet,C:FiniteSet,B:FiniteSet)(f:A->C,g:B->C):(Pullback:FiniteSet,Pullback->A,Pullback->B,(X:FiniteSet,X->A,X->B)->(X->Pullback))==
--          This uses Theorem 4.5 in McLarty
        import from o(FiniteSet,AB,A,C),o(FiniteSet,AB,B,C)
	
	(AB:FiniteSet,pa:AB->A,pb:AB->B,product:(X:FiniteSet)->(X->A,X->B)->(X->AB)) == Product(A,B)
	(E:FiniteSet,e:E->AB) == Equalizer(AB,C,f ** pa, g ** pb)
	
	import from o(FiniteSet,E,AB,A),o(FiniteSet,E,AB,B)
	(E,pa**e,pb**e,error " ")
    
